
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.29.3',
    'branch_nick': 'release-1.29-LTS',
    'revision_id': '82c0b9ac28dad63d87bba82622cc84327d74c8bd',
    'revno': 11976
}
